// rewritten by Zheng Gengbin
// analyse:
//  analyse_alignment()
//  analyse_realignment()
//  analyse_distribute()
//  analyse_redistribute()
//  analyse_dynamic()
//  inherit
//  memory
//  incoresize
//  outofcore
//  major

#include <stdio.h>
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
extern "C" void exit(...);
#endif

#include "compilerApi.h"

#include "anls_att.h"

extern void add_var_dec(SgStatement *, SgType *, SgSymbol *);
extern int getreflevel(SgSymbol * index, SgExpression * expr);
extern SgExpression * getlinerexp(int i, SgSymbol * symb, SgExpression * expr);

extern SgStatement * wrap_subroutine(SgStatement *s );

void analyse_alignment(SgStatement *);
void show_alignment(SgStatement *);

void add_dimention(Alignment_def * adef, SgExpression * expr)
{
  SgExpression * alignleft = expr->operand(1);
  SgArrayRefExp * alignright = (SgArrayRefExp *)(expr->operand(2)->operand(2));
  
	for (SgExpression *alignl = alignleft ; alignl ; alignl=alignl->rhs())
	{
		SgExpression *source = alignl->lhs();
		ASSERT (source->variant() == VAR_REF);
		SgSymbol   *dummy = source->symbol();
	
		int i=0;
		SgExpression *with = alignright->lhs();
		for ( ; with ; with=with->rhs(),i++ )
			if (with->lhs()->IsSymbolInExpression(*dummy)) break;
		// if not found, then must collapsed
		if (with == NULL) 
		{
			adef->setAlignPair(-1);
    			adef->add_dimention_info(NULL, NULL);
		}
		else 
		{
			SgExpression *node = with->lhs();
/*
		s=getlinerexp(1, index, alignright->subscript(i));
    		o=getlinerexp(0, index, alignright->subscript(i));
*/
    			SgExpression * s=node->lhs()->lhs();
    			SgExpression * o=node->rhs();
			adef->setAlignPair(i);
    			adef->add_dimention_info(s, o);
		}
	}
	adef->setAlignWithDim(alignright->numberOfSubscripts());
}

// deal_Alignment_array add the source and target of alignment to the program header

void deal_Alignment_array(SgStatement * s, SgExpression * expr1, SgExpression * expr2)
{
  Alignment_def * adef =new Alignment_def;

  SgSymbol * srcsymb = expr1->operand(1)->symbol();
  SgSymbol * tarsymb = expr2->operand(1)->operand(2)->operand(2)->symbol();

  // add the alignment source and target;
  adef->setsymbol(srcsymb, tarsymb);

  // the following operand(1) should be an ALIGN_OP;
  add_dimention(adef, expr2->operand(1)); 

  // add align def to the program header and source symbol;
  s->controlParent()->addAttribute(Align_Def, (void *) adef, sizeof(Alignment_def));
  adef->getfromsymbol()->addAttribute(Align_Def, (void *) adef, sizeof(Alignment_def));     

  // align array to non-template array( dummy )
  SgType *type = tarsymb->type();
  if (type->hasBaseType()) type = type->baseType();
  if (type->variant() != T_TEMPLATE)
  {
if (!silentFlag)
printf("Symb:'%s' has unknown bound.\n",srcsymb->identifier());
     srcsymb->addAttribute(UNBOUND_ARRAY, NULL, 0);
  }
}

void analyse_alignment(SgStatement * s)
{
  deal_Alignment_array(s, s->expr(0), s->expr(2));
}

// for ReAlign Statement
// deal_Alignment_array add the source and target of alignment to the program header
// by Zheng Gengbin
void deal_ReAlignment_array(SgStatement * s, SgExpression * expr1, SgExpression * expr2)
{
  Alignment_def * adef =new Alignment_def;

  SgSymbol * srcsymb = expr1->operand(1)->symbol();
  SgSymbol * tarsymb = expr2->operand(1)->operand(2)->operand(2)->symbol();

  // add the alignment source and target;
  adef->setsymbol(srcsymb, tarsymb);

  // the following operand(1) should be an ALIGN_OP;
  add_dimention(adef, expr2->operand(1)); 

  // add align def to the program header and source symbol;
  s->addAttribute(Align_Def, (void *) adef, sizeof(Alignment_def));
//  adef->getfromsymbol()->addAttribute(Align_Def, (void *) adef, sizeof(Alignment_def));     
}

void analyse_realignment(SgStatement * s)
{
  deal_ReAlignment_array(s, s->expr(0), s->expr(2));
}

void show_alignment(SgStatement * s)
{
  for (int i=0; i<s->numberOfAttributes(Align_Def); i++) {
    ((Alignment_def *) ((s->getAttribute(i, Align_Def))->getAttributeData()) ) ->show();
  };
}

//------------------------------------;

void analyse_distribute(SgStatement *);
void show_distribute(SgStatement *);

void add_dist_dimention(Distribute_def * ddef, SgExpression * expr)
{
  int s;

  //printf("----%d\n", expr->id());
  switch (expr->operand(1)->variant()) {
  case STAR_OP:
    s=NOP; 
    break;
  case BLOCK_OP:
    s=BLOCK;
    break;
  case CYCLIC_OP:
    s=CYCLIC;
    break;
  default:
    printf("Wrong definition\n");
    break;    
  };

  ddef->add_dimention_info(s);

  if (expr->operand(2))
    add_dist_dimention(ddef, expr->operand(2));

}

// deal_Distribute_array add the source and target of distribute to the program header

void deal_Distribute_array(SgStatement * s, SgExpression * expr1, SgExpression * expr2)
{
  ASSERT (isSgDistributeStmt(s));
  SgDistributeStmt *dist_stmt = isSgDistributeStmt(s);

  Distribute_def * ddef = new Distribute_def;

    // set subgroup procs
  SgExpression *procExp = expr2->lhs()->rhs()->rhs();
  ddef->setProcs(procExp);

    // add the alignment source and target.
  SgSymbol *fromSym = expr1->lhs()->symbol();
  SgSymbol *toSym = procExp->symbol();
  ddef->setsymbol(fromSym, toSym);
  
    // decide the processor group name
  SgSymbol *procgrd ;
  if (procExp->variant() == VAR_REF)
	procgrd = s->getSymbolInCurrentScope(mixstring, "grp", toSym->identifier());
  else
	procgrd = s->getSymbolInCurrentScope(mixstring, "subgrp", fromSym->identifier());
  ddef->setgrp(procgrd);

    // add dimention info, look for the right position.
  add_dist_dimention(ddef, expr2->lhs()->lhs()->rhs());
  if (dist_stmt->is_descriptive())  ddef->set_descriptive(1);

    // add attributes
  s->controlParent()->addAttribute(Dist_Def, (void *) ddef, sizeof(Distribute_def));

    // also add to the from symbol
  ddef->getfromsymbol()->addAttribute(Dist_Def, (void *) ddef, sizeof(Distribute_def));

  // add distribute def to the program header and source symbol;    by zheng.gb
  // here a symbol is directly distributed without align 
  if (  ddef->getfromsymbol()->numberOfAttributes(Temp_Def) == 0 )
  {
  	s->controlParent()->addAttribute(DDist_Def, (void *) ddef, sizeof(Distribute_def));
  }
}

void analyse_distribute(SgStatement * s)
{
  deal_Distribute_array(s, s->expr(0), s->expr(2));
}

// for ReDistribute Statement
// deal_Distribute_array add the source and target of distribute to the program header
// add by Zheng Gengbin
void deal_ReDistribute_array(SgStatement * s, SgExpression * expr1, SgExpression * expr2)
{
  SgReDistributeStmt *dist_stmt = isSgReDistributeStmt(s);
  ASSERT (dist_stmt);

  Distribute_def * ddef = new Distribute_def;

  SgExpression *procExp = expr2->lhs()->rhs()->rhs();
  ddef->setProcs(procExp);

  // add the alignment source and target.
  SgSymbol *fromSym = expr1->lhs()->symbol();
  SgSymbol *toSym = procExp->symbol();
  ddef->setsymbol(fromSym, toSym);
  
   // decide the processor group name
  SgSymbol *procgrd ;
  if (procExp->variant() == VAR_REF)
	procgrd = s->getSymbolInCurrentScope(mixstring, "grp", toSym->identifier());
  else
	procgrd = s->getSymbolInCurrentScope(mixstring, "subgrp", fromSym->identifier());
  ddef->setgrp(procgrd);

    // add dimention info, look for the right position.
  add_dist_dimention(ddef, expr2->lhs()->lhs()->rhs());
//  if (dist_stmt->is_descriptive())  ddef->set_descriptive(1);

  s->addAttribute(Dist_Def, (void *) ddef, sizeof(Distribute_def));
}

void analyse_redistribute(SgStatement * s)
{
  deal_ReDistribute_array(s, s->expr(0), s->expr(2));
}

void show_distribute(SgStatement * s)
{
  for (int i=0; i<s->numberOfAttributes(Dist_Def); i++) {
    ((Distribute_def *) s->attributeValue(i, Dist_Def)) ->show();
  };
}

//
void analyse_dynamic(SgStatement *s)
{
   SgExpression *var_list = s->expr(0);
   for ( ; var_list; var_list=var_list->rhs())
   {
       SgSymbol *symb = var_list->lhs()->symbol();
       symb->addAttribute(Dynamic_Def, NULL, 0);
   }
}

//-------------------------;

/* check this declaration is legal , called by legal_inconstant_bound() */
void pass_bound(SgStatement *s, SgExpression *exp, int &leg)
{
    if (exp) {
	switch (exp->variant())
	{
	case VAR_REF:
		if (is_dummy(s, exp->symbol()) != 1) leg=0;
		break;
	case FUNC_CALL:
	case INTRIN_FUNC_CALL:
		leg=0;
		break;
	default:
	     pass_bound(s, exp->lhs(), leg);
	     pass_bound(s, exp->rhs(), leg);
	}
    }
}


/* check this declaration is legal */
int legal_inconstant_bound(SgStatement *s, SgSymbol *arrayname, SgExpression *uppervalue)
{
    if (uppervalue->isInteger()) return 1;
    if (is_dummy(s,arrayname) == 1) return 1;
    int leg = 1;
    pass_bound(s,uppervalue, leg);
    return leg;
}

void add_align_array(SgStatement * s)
{
  for (int i=0; i<s->numberOfAttributes(Align_Def); i++) {
    Alignment_def * adef = (Alignment_def *)s->attributeValue(i,Align_Def);

    SgSymbol * arrayname = adef->getfromsymbol();
    SgSymbol * alignee = adef->gettosymbol();

    SgType * arraytype = arrayname->type();

    //also record this symbol;
    //adef->setlocal(arrayname);
    
    SgExpression * array = new SgArrayRefExp(*arrayname);
    SgExpression * tail = array;

    SgExpression *uppervalue=NULL;

    if(is_out_of_core(arrayname))  {
	uppervalue = get_in_core_size(s);
    } 
    else 	//  the variable is aligned to a dummy
    if (alignee->numberOfAttributes(Temp_Def) == 0)
    {
//	printf("I don't know how to deal with a non-template symbol :%s.",arrayname->identifier());
//	ASSERT(adef->gettosymbol()->numberOfAttributes(Temp_Def)  != 0);
// 	define this variable as it was.
/*
	SgArrayType *arraytype = isSgArrayType(arrayname->type());
	ASSERT(arraytype);
	uppervalue = arraytype->getSizeExp();
	ASSERT(uppervalue);
*/
	// let allocate memory for it
	continue;
    }
    else if (is_dynamic(arrayname))
    {
	  // DYNAMIC will allocate memory
	continue;
    }
    else  {
      Shadow_def * sdef = NULL;
	// search for Global Shadow
      if (arrayname->numberOfAttributes(Shad_Def)!=0) {
        for (int k=0; k<arrayname->numberOfAttributes(Shad_Def); k++) {
	   Shadow_def *gsdef = (Shadow_def *)arrayname->attributeValue(k,Shad_Def);
	   if (gsdef->getTag() == GNEEDED) {
		sdef = gsdef;
	  	break;
	   }
        } // for
      }

      //  alow align permutation
      Template_def *tdef = (Template_def*)alignee->attributeValue(0, Temp_Def);
      uppervalue = new SgValueExp(1);
      for (int j=0; j<adef->getDim(); j++) {
	SgExpression *length=NULL ;
	if ( adef->getAlignPair(j) != -1)
		length = tdef->getlocalsize(adef->getAlignPair(j));
	else {		// use the array size of dimension j
		SgExpression *ub = get_ubound(arrayname,j);
		SgExpression *lb = get_lbound(arrayname,j);
		length = &(*ub - *lb + *new SgValueExp(1));
	}
	ASSERT(length);
	if ( sdef != NULL )
	{
	    // use j or the alignPair ???  zheng.gb
	  SgExpression *l=new SgValueExp(sdef->getLower(j));
	  SgExpression *u=new SgValueExp(sdef->getUpper(j));
	  length = &( *length + *l + *u);
	}
	uppervalue = &(*uppervalue * *length);
	uppervalue = calculateExp(uppervalue);
	if (uppervalue->isInteger())
	   ASSERT(uppervalue->valueInteger()>0);
      };

    } // end of else

    // to here we get the uppervalue
    // create the var decl statement to declare it.
    SgExpression * c1 = new SgValueExp(0);
    SgExpression * c2 = &( *uppervalue - *new SgValueExp(1));
    c2 = calculateExp(c2);

    SgExpression * range = new SgExpression(RANGE_OP, c1, c2, NULL);
    SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);
    SgExpression * ubexpr = new SgExpression(EXPR_LIST, ddot, NULL, NULL);

    tail->setLhs(ubexpr);
    tail=tail->operand(2);

    SgType * atype = arraytype->internalBaseType();
    SgExpression * arraytypeExp = new SgExpression(TYPE_OP, NULL, NULL, NULL, atype);
    // check kind
    if (isSgArrayType(arraytype)) {
	SgType *tp = isSgArrayType(arraytype)->directBaseType();
	if (isSgKindType(tp)){
		SgExpression *kindExp = isSgKindType(tp)->kindVal();
		arraytypeExp->setLhs(kindExp);
	} 
    }

    SgStatement * declarray = new SgStatement(VAR_DECL, NULL, NULL, array, arraytypeExp, NULL);
  
    // some fortran compiler can not accept 
#ifndef FORTRAN_DYNAMIC_ARRAY_SIZE
    if (!uppervalue->isInteger())
	s->addAttribute(INCONST_ARRAY, array, sizeof(array));
    else
#endif
//    if (is_dummy(s,arrayname) == 0  && !uppervalue->isInteger())

      // wrap this array later
    if (legal_inconstant_bound(s, arrayname, uppervalue) == 0)
	s->addAttribute(INCONST_ARRAY, array, sizeof(array));
    else 
       	s->childList1(0)->insertStmtBefore(*declarray, *s);

    // update stack_size
    if (isSgArrayType(arrayname->type()) && stack_size>0) {
	SgArrayType *type = isSgArrayType(arrayname->type());
	int arraysize = type->getSize();
	switch (arrayname->type()->baseType()->variant()) {
	case T_INT:
	case T_FLOAT:
	case T_BOOL:
		arraysize *= 4; break;
	case T_DOUBLE:
	case T_COMPLEX:
	case T_DCOMPLEX:
		arraysize *= 8; break;
	case T_CHAR:
        case T_BYTE:
		break;
	}
	if (arraysize>0 ) 
		stack_size += arraysize;
    }

  };	// end of for Align_Def

  // wrap subroutine by adding more un-determined argument
  if (s->numberOfAttributes(INCONST_ARRAY) != 0)
  {
extern void add_dummy_array(SgStatement *);
extern void add_definition(SgStatement *);
extern void add_global_common(SgStatement *s, SgStatement* lastdec);
  	SgStatement *newproc = wrap_subroutine(s);
	if (in_main_program == on) {
	  add_definition(newproc);
	  add_dummy_array(s);
	}
	else {	// in sub
	  add_global_common(newproc, newproc->lastDeclaration());
	  add_dummy_array(newproc);
	}
  }
}

//---------------------;

extern void add_var_dec(SgStatement *, SgType *, SgSymbol *);

void add_temp_base(SgStatement * s)
{
}

// ---------------------;

void delete_align_array(SgStatement * s)
{
  if (s->variant()==VAR_DECL) {
    SgDeclarationStatement * declstmt;
    declstmt = (SgDeclarationStatement *) s ;

    SgExpression **  distvar = new SgExpression *[declstmt->numberOfVars()];
    int counter=0;

    int i;

    for (i=0; i<declstmt->numberOfVars(); i++) {
      SgSymbol * varsym = declstmt->var(i)->symbol();
      //      printf("The symbol is --- %s\n", varsym->identifier());
      if(varsym->numberOfAttributes(Align_Def)!=0) {
	distvar[counter]=declstmt->var(i);
	counter++;
      }
    }

    for (i=0; i<counter; i++) {
      declstmt->deleteTheVar(* (distvar[i]));
    }

    if (declstmt->numberOfVars()==0)
      declstmt->deleteStmt();

    delete [] distvar;
  };
}

// process var decl statements and delete the allocatable and pointer variable
void delete_allocatable_array(SgStatement * s)
{
  if (s->variant()==VAR_DECL) {
    SgDeclarationStatement * declstmt;
    declstmt = (SgDeclarationStatement *) s ;

    SgExpression **  distvar = new SgExpression *[declstmt->numberOfVars()];
    int counter=0;

    int i;

    for (i=0; i<declstmt->numberOfVars(); i++) {
      SgSymbol * varsym = declstmt->var(i)->symbol();
      //      printf("The symbol is --- %s\n", varsym->identifier());
      if(is_allocatable(varsym) || is_pointer(varsym)) {
	distvar[counter]=declstmt->var(i);
	counter++;
      }
    }

    for (i=0; i<counter; i++) {
      declstmt->deleteTheVar(* (distvar[i]));
    }

    if (declstmt->numberOfVars()==0)
      declstmt->deleteStmt();

    delete [] distvar;
  };
}

//-------------------;
void add_dummy_array(SgStatement * s)
{
  SgProcHedrStmt *  prochedr = (SgProcHedrStmt *) s;
  int org_numberOfParameters = prochedr->numberOfParameters();

  for (int i=0; i<org_numberOfParameters; i++) {
    SgSymbol * parameter = prochedr->parameter(i);
    if (parameter->type()->variant()==T_ARRAY) {
      SgArrayType * arraytype = (SgArrayType *) parameter->type();

      SgArrayRefExp * declist = new SgArrayRefExp ( * parameter);
// by zheng.gb 1997/9/10		?? only one 0:* is enough ?
//      if (parameter->numberOfAttributes(Inherit_Def)!=0) {
//	for (int dimI = 0; dimI < arraytype->dimension(); dimI++) {
	  SgValueExp * zero = new SgValueExp(0);
	  SgExpression * star = new SgExpression (STAR_RANGE, NULL, NULL, NULL);
	  SgExpression * ddot = new SgExpression (DDOT, zero, star, NULL);
	  
	  declist->addSubscript( *ddot);
//	}	// end of for
//      }
//      else
//	printf("add_dummy_array:not supported yet!\n");

      
      SgVarDeclStmt * decstmt = new SgVarDeclStmt(*declist, * arraytype->baseType());

      s->childList1(0)->insertStmtBefore(*decstmt, *s);
    }
  }

}


// ---------------------;

void delete_dummy_array(SgStatement * s)
{
  if (s->variant()==VAR_DECL) {
    SgDeclarationStatement * declstmt;
    declstmt = (SgDeclarationStatement *) s ;

    SgExpression *  distvar[10];
    int counter=0;

    int i;

    for (i=0; i<declstmt->numberOfVars(); i++) {
      SgSymbol * varsym = declstmt->var(i)->symbol();
      //      printf("The symbol is --- %s\n", varsym->identifier());
      if(varsym->numberOfAttributes(DUMP_Def)!=0) {
	if(varsym->type()->variant()==T_ARRAY) {
	  distvar[counter]=declstmt->var(i);
	  counter++;
	}
      }
    }

    for (i=0; i<counter; i++) {
      declstmt->deleteTheVar(* (distvar[i]));
    }

    if (declstmt->numberOfVars()==0)
      declstmt->deleteStmt();
  };
}

//----------------;

void analyse_inherit(SgStatement * s)
{
  SgInheritStmt * inherit = (SgInheritStmt *) s;

  for (int i=0; i<inherit->numberOfVars(); i++)
    inherit->var(i)->symbol()->addAttribute(Inherit_Def, NULL, 0);
}

Shadow_def * deal_shadow_array(SgStatement *s,SgExprListExp *explist)
{
   Shadow_def *sdef = new Shadow_def(DEFINED);
   for (int i=0; i<explist->length(); i++) 
   {
	SgTripletExp *ddot = isSgTripletExp(explist->elem(i));
	SgExpression *lb = ddot->item(0);
	SgExpression *ub = ddot->item(1);
	if (lb == NULL) lb = ub->copyPtr();
	ASSERT(lb->isInteger() && ub->isInteger());
	sdef->setLower(i,lb->valueInteger());
	sdef->setUpper(i,ub->valueInteger());
   }
   return sdef;
}

void analyse_shadow(SgStatement *s)
{
   SgExprListExp *varlist = (SgExprListExp *)s->expr(0);
   for (int  i=0; i<varlist->length(); i++) 
   {
	SgSymbol *symb = varlist->elem(i)->symbol();
	Shadow_def *sdef = deal_shadow_array(s,(SgExprListExp *)s->expr(2)->lhs());
	symb->addAttribute(Shad_Def, sdef, sizeof(sdef));
   }
}

void analyse_outofcore(SgStatement * s)
{
  SgInheritStmt * inherit = (SgInheritStmt *) s;

  for (int i=0; i<inherit->numberOfVars(); i++)
   { 
	inherit->var(i)->symbol()->addAttribute(Outofcore_Def, NULL, 0);
   }
}

// program argument overwrite this function
void analyse_incoresize(SgStatement * s)
{
   SgExpression *size = s->expr(0);
   ASSERT(size->isInteger());
   IN_CORE_SIZE = size->valueInteger(); 
}


// program argument overwrite this function
void analyse_memory(SgStatement * s)
{
     // negative stack_size means program -M specified.
   if (stack_size < 0) return;
   SgExpression *size = s->expr(0);
   ASSERT(size->isInteger());
   stack_size = - size->valueInteger(); 
}


// parse major definition
void analyse_major(SgStatement * s)
{
  SgInheritStmt * inherit = (SgInheritStmt *) s;

  for (int i=0; i<inherit->numberOfVars(); i++)
   { 
	SgExpression *lhs = inherit->var(i);
	SgSymbol *symb = lhs->symbol();
	switch (lhs->variant()) {
	case COLUMN_OP:
		symb->addAttribute(Major_Def, (void *)0, sizeof(int));
		break;
	case ROW_OP:
		symb->addAttribute(Major_Def, (void *)1, sizeof(int));
		break;
	default:
		ASSERT(0);
	}
   }
}


/*********************************************
	COMMON (not finished)
**********************************************/

void analyse_common(SgStatement *s)
{
    SgExpression * comm_list =  s->expr(0);
    for (; comm_list; comm_list=comm_list->rhs())
    {
	SgSymbol *commName = comm_list->symbol();
	SgExpression *exprlist = comm_list->lhs();
	for ( ; exprlist; exprlist=exprlist->rhs() )
	{
		SgExpression *expr = exprlist->lhs();
		SgSymbol *symb = expr->symbol();
		symb->addAttribute(Common_Def, expr, sizeof(expr));
	}
    }
}


